local Direction = require('Direction')
local LocalInput = require('Class').create(nil, 'LocalInput')
local ControllerComponent = require('Components.ControllerComponent')
local EntitySet = require('EntitySet')
local SpriteSheet = require('SpriteSheet')

LocalInput.KEY_PROMPTS = {
    [string.byte('0')] = '0',
    [string.byte('1')] = '1',
    [string.byte('2')] = '2',
    [string.byte('3')] = '3',
    [string.byte('4')] = '4',
    [string.byte('5')] = '5',
    [string.byte('6')] = '6',
    [string.byte('7')] = '7',
    [string.byte('8')] = '8',
    [string.byte('9')] = '9',
    [string.byte('a')] = 'A',
    [string.byte('b')] = 'B',
    [string.byte('c')] = 'C',
    [string.byte('d')] = 'D',
    [string.byte('e')] = 'E',
    [string.byte('f')] = 'F',
    [string.byte('g')] = 'G',
    [string.byte('h')] = 'H',
    [string.byte('i')] = 'I',
    [string.byte('j')] = 'J',
    [string.byte('k')] = 'K',
    [string.byte('l')] = 'L',
    [string.byte('m')] = 'M',
    [string.byte('n')] = 'N',
    [string.byte('o')] = 'O',
    [string.byte('p')] = 'P',
    [string.byte('q')] = 'Q',
    [string.byte('r')] = 'R',
    [string.byte('s')] = 'S',
    [string.byte('t')] = 'T',
    [string.byte('u')] = 'U',
    [string.byte('v')] = 'V',
    [string.byte('w')] = 'W',
    [string.byte('x')] = 'X',
    [string.byte('y')] = 'Y',
    [string.byte('z')] = 'Z',
    [MOAIKeyboardSensor.ALT] = 'Alt',
    [string.byte('`')] = 'Apostrophe',
    [string.byte('\\')] = 'BackSlash',
    [8] = 'Backspace',
    [string.byte(',')] = 'Comma',
    [MOAIKeyboardSensor.CONTROL] = 'Ctrl',
    [127] = 'Del',
    [MOAIKeyboardSensor.UP] = 'Up',
    [MOAIKeyboardSensor.DOWN] = 'Down',
    [MOAIKeyboardSensor.LEFT] = 'Left',
    [MOAIKeyboardSensor.RIGHT] = 'Right',
    [MOAIKeyboardSensor.END] = 'End',
    [10] = 'Enter',
    [13] = 'Enter',
    [string.byte('=')] = 'Equal',
    [27] = 'Esc',
    [MOAIKeyboardSensor.F1] = 'F1',
    [MOAIKeyboardSensor.F2] = 'F2',
    [MOAIKeyboardSensor.F3] = 'F3',
    [MOAIKeyboardSensor.F4] = 'F4',
    [MOAIKeyboardSensor.F5] = 'F5',
    [MOAIKeyboardSensor.F6] = 'F6',
    [MOAIKeyboardSensor.F7] = 'F7',
    [MOAIKeyboardSensor.F8] = 'F8',
    [MOAIKeyboardSensor.F9] = 'F9',
    [MOAIKeyboardSensor.F10] = 'F10',
    [MOAIKeyboardSensor.F11] = 'F11',
    [MOAIKeyboardSensor.F12] = 'F12',
    [MOAIKeyboardSensor.HOME] = 'Home',
    [MOAIKeyboardSensor.INSERT] = 'Insert',
    [string.byte('[')] = 'LeftBracket',
    [string.byte(']')] = 'RightBracket',
    [string.byte('-')] = 'Minus',
    [MOAIKeyboardSensor.PAGE_DOWN] = 'PageDown',
    [MOAIKeyboardSensor.PAGE_UP] = 'PageUp',
    [string.byte('.')] = 'Period',
    [string.byte(';')] = 'Semicolon',
    [MOAIKeyboardSensor.SHIFT] = 'Shift',
    [string.byte('/')] = 'Slash',
    [string.byte(' ')] = 'Space',
    [string.byte('\t')] = 'Tab',
    [string.byte('~')] = 'Tilde',
}

function LocalInput:init (userState)
    self.textInput = ""
    self.keys = {}
    self.keyBindings = {}
    self.entitySet = EntitySet.new()
    MOAIInputMgr.device.keyboard:setCallback(function (keyCode, down) self:onKey(keyCode, down) end)
    MOAIInputMgr.device.keyboard:setTextCallback(function (text) self:onText(text) end)

    self.userState = userState
    self.userState.controlsChangeEvent:register(self.onControlsChangeEvent, self)
    self:onControlsChangeEvent()
    
    self.gamepadPrompts = {
        ['Up'] = 'DPad_Up',
        ['Left'] = 'DPad_Left',
        ['Down'] = 'DPad_Down',
        ['Right'] = 'DPad_Right',
        ['Interact'] = 'X',
        ['Select'] = 'A',
        ['UsePrimary'] = 'A',
        ['Use1'] = 'B',
        ['Use2'] = 'Y',
        ['Use3'] = 'LB',
        ['Use4'] = 'RB',
        ['OpenTabs'] = 'Back',
        ['OpenMenu'] = 'Start',
        ['Dismiss'] = 'B',
    }
end

function LocalInput:onControlsChangeEvent ()
    self.keyPrompts = {}
    self.pendingBindings = {}
    for motive, keys in pairs(self.userState.data.controls.KEY_BINDINGS) do
        local primaryKey = keys[1]
        self.keyPrompts[motive] = LocalInput.KEY_PROMPTS[primaryKey]
        self.pendingBindings[motive] = keys
    end
    
    for entity in pairs(self.entitySet.entities) do
        local controller = entity:get(ControllerComponent)
        if self.lastMode == 'gamepad' then
            self:setGamepadMode(controller)
        else
            self:setKeyboardMode(controller)
        end
    end
end

function LocalInput:driveEntity (entity)
    self.entitySet:addEntity(entity)
    
    local controller = entity:get(ControllerComponent)
    if self.lastMode == 'gamepad' then
        self:setGamepadMode(controller)
    else
        self:setKeyboardMode(controller)
    end
    controller:mapMotive('Up', 'MoveN')
    controller:mapMotive('Left', 'MoveW')
    controller:mapMotive('Down', 'MoveS')
    controller:mapMotive('Right', 'MoveE')
    
    -- make sure motives that are already held do not get registered as new motives
    self:tickController(controller)
    controller:clearNewMotives()
end

function LocalInput:tick ()
    for entity in pairs(self.entitySet.entities) do
        local controller = entity:get(ControllerComponent)
        self:tickController(controller)
    end
    if self.textInput ~= "" then
        self.textInput = ""
    end
    if #self.keys ~= 0 then
        self.keys = {}
    end
end

function LocalInput:onKey (keyCode, down)
    if down then
        table.insert(self.keys, keyCode)
    end
end

function LocalInput:onText (text)
    self.textInput = self.textInput .. text
end

function LocalInput:tickController (controller)
    controller:clearNewMotives()

    -- First aggregate all of the keyboard motives
    local motives = {}
    local hasKeyboardMotive = false
    for motive, keys in pairs(self.keyBindings) do
        local motiveOn = false
        for index, key in ipairs(keys) do
            if MOAIInputMgr.device.keyboard:keyIsDown(key) then
                hasKeyboardMotive = true
                motiveOn = true
                break
            end
        end
        motives[motive] = motiveOn
    end
    
    -- Check to see if we have pending bindings that we should activate
    local activatedBindings = {}
    for motive, keys in pairs(self.pendingBindings) do
        -- We can only activate a binding if the existing binding is off
        -- and the new binding is off as well
        if not motives[motive] then
            local pendingReady = true
            local keysCopy = {}
            for index, key in ipairs(keys) do
                if MOAIInputMgr.device.keyboard:keyIsDown(key) then
                    pendingReady = false
                    break
                else
                    table.insert(keysCopy, key)
                end
            end
            if pendingReady then
                self.keyBindings[motive] = keysCopy
                table.insert(activatedBindings, motive)
            end
        end
    end
    for index, motive in ipairs(activatedBindings) do
        self.pendingBindings[motive] = nil
    end
    
    -- Then aggregate all of the controller motives
    local hasGamepadMotive = false

    -- Look for a direction vector
    local vx, vy = MOAIInputMgr.device.gamepadStickLeft:getVector()
    
    -- Deadzone fix by daf based on example code from
    -- http://www.third-helix.com/2013/04/12/doing-thumbstick-dead-zones-right.html
    local deadzone = 0.25 -- deadzone value bettwen 0 and 1
    if math.abs(vx) < deadzone then
        vx = 0
    end
    if math.abs(vy) < deadzone then
        vy = 0
    end
    -- End of Deadzone fix changes
    
    local dir = nil
    if vx ~= 0 or vy ~= 0 then
        local mag = math.sqrt(vx * vx + vy * vy)
        dir = Direction.vectorToDir(vx, -vy, true)
        vx, vy = Direction.dirToVector(dir)
        vx, vy = vx * mag, vy * mag
        hasGamepadMotive = true
    end
    controller:setDirectionVector(vx, vy)
    
    local function setGamepadMotive(motive, on)
        motives[motive] = motives[motive] or on
        if on then
            hasGamepadMotive = true
        end
    end
    setGamepadMotive('Up', MOAIInputMgr.device.gamepadButtonDPadUp:isDown() or (dir == Direction.DIR_N))
    setGamepadMotive('Left', MOAIInputMgr.device.gamepadButtonDPadLeft:isDown() or (dir == Direction.DIR_W))
    setGamepadMotive('Down', MOAIInputMgr.device.gamepadButtonDPadDown:isDown() or (dir == Direction.DIR_S))
    setGamepadMotive('Right', MOAIInputMgr.device.gamepadButtonDPadRight:isDown() or (dir == Direction.DIR_E))
    setGamepadMotive('Interact', MOAIInputMgr.device.gamepadButtonX:down())
    setGamepadMotive('Select', MOAIInputMgr.device.gamepadButtonA:down())    
    setGamepadMotive('UsePrimary', MOAIInputMgr.device.gamepadButtonA:down())
    setGamepadMotive('Use1', MOAIInputMgr.device.gamepadButtonB:isDown())
    setGamepadMotive('Use2', MOAIInputMgr.device.gamepadButtonY:isDown())
    setGamepadMotive('Use3', MOAIInputMgr.device.gamepadButtonLShoulder:isDown())
    setGamepadMotive('Use4', MOAIInputMgr.device.gamepadButtonRShoulder:isDown())
    setGamepadMotive('OpenTabs', MOAIInputMgr.device.gamepadButtonBack:down())
    setGamepadMotive('OpenMenu', MOAIInputMgr.device.gamepadButtonStart:down())
    setGamepadMotive('Dismiss', MOAIInputMgr.device.gamepadButtonB:down())
    
    -- Assign all the motives
    for motive, on in pairs(motives) do
        controller:setMotive(motive, on)
    end
    
    -- Set the current input mode
    if hasKeyboardMotive then
        self:setKeyboardMode(controller)
    elseif hasGamepadMotive then
        self:setGamepadMode(controller)
    end
    
    -- Send any keyboard as text input
    controller:setTextInput(self.textInput)
    if #self.keys ~= 0 then
        controller:setKeys(self.keys)
    else
        controller:clearKeys()
    end
end

function LocalInput:setKeyboardMode (controller)
    self.lastMode = 'keyboard'
    controller:setMode(SpriteSheet.load('UI/Prompts/Keyboard/Keyboard'), self.keyPrompts)
end

function LocalInput:setGamepadMode (controller)
    self.lastMode = 'gamepad'
    controller:setMode(SpriteSheet.load('UI/Prompts/360/360'), self.gamepadPrompts)
end

return LocalInput
